# -*- coding: utf-8 -*-
import collections
from watson.form.fields import File, Hidden, Definition
from watson.html.elements import TagMixin, flatten_attributes
from watson.common.contextmanagers import suppress
from watson.common.datastructures import MultiDict
from watson.common.decorators import cached_property
from watson.common.imports import get_qualified_name


IGNORED_ATTRIBUTES = ('fields', 'data', 'raw_data', 'errors')


class FieldDescriptor(object):
    """Allow set/get access to the field value via Form.field_name

    Fields accessed via a Form object will provide access directly to
    set and get the value of the field. When the field is being rendered
    in a template, the set/get will provide access to the field object
    itself.
    Access to the field prior to rendering can be made via Form.fields[name]
    where name is the attribute name of the field.
    """
    def __init__(self, name):
        self.name = name

    def __get__(self, instance, owner):
        field = instance.fields[self.name]
        if instance._rendering:
            return field
        return field.value

    def __set__(self, instance, value):
        instance.fields[self.name].value = value


class FormMeta(type):
    """Assigns the FieldDescriptor objects to the Form object.
    """
    def __init__(cls, name, bases, attrs):
        fields = []
        for field_name in dir(cls):
            if (not field_name.startswith('_') and
                    field_name not in IGNORED_ATTRIBUTES):
                field = getattr(cls, field_name)
                if isinstance(field, Definition):
                    fields.append((field_name, field))
        fields.sort(key=lambda pair: pair[1].count)
        cls._defined_fields = collections.OrderedDict(fields)

    def __call__(cls, *args, **kwargs):
        return type.__call__(cls, *args, **kwargs)


class Form(TagMixin, metaclass=FormMeta):
    """Declarative HTML <form> management.

    Example:

    .. code-block:: python

        from watson.form import fields
        class MyForm(Form):
            text = fields.Text(name='text', label='My TextField')
            another = fields.Checkbox(name='another[]')

        form = MyForm('my_form')
        form.text = 'Something'

        # in view
        {% form.open() %}   # <form name="my_form">
        {% form.text %}     # <input name="text" type="text" value="Something" />
        {% form.text.render_with_label() %} # <label for="text">My TextField</label><input id="text" name="text" type="text" value="Something" />
        {% form.another %}  # <input name="another[]" />
        {% form.close() %}  # </form>
    """
    _fields = None
    _mapped_fields = None
    _rendering = False
    _ignored_bound_fields = None
    _bound_object = None
    _bound_object_mapping = None
    _validated = False

    def __init__(self, name=None, method='post',
                 action=None, detect_multipart=True, validators=None,
                 values_provider=None, **kwargs):
        """Inititalize the form and set some default attributes.

        Args:
            name (string): the name of the form
            method (string): the http method to use
            action (string): the url to submit the form to
            detect_multipart (boolean): automatically set multipart/form-data
        """
        self._ignored_bound_fields = []
        self.validators = validators or []
        self._process_tag_attributes(name, method, action, kwargs)
        self._change_http_method(method)
        self._set_values_provider(values_provider)
        self._detect_multipart(detect_multipart)

    def _process_tag_attributes(self, name, method, action, kwargs):
        if '_class' in kwargs:
            kwargs['class'] = kwargs.get('_class')
            del kwargs['_class']
        self.attributes = collections.ChainMap({
            'name': name or self.__class__.__name__,
            'method': method.lower(),
            'action': action or '/',
            'enctype':
            kwargs.get('enctype', 'application/x-www-form-urlencoded')
        }, kwargs)

    def _change_http_method(self, method):
        if self.method not in ('get', 'post'):
            field_name = 'http_request_method'
            self.attributes['method'] = 'post'
            self._add_field(field_name, Hidden(
                name='HTTP_REQUEST_METHOD',
                value=method.upper(),
                definition=False))

    def _add_field(self, field_name, field):
        self.fields[field_name] = field
        setattr(self.__class__, field_name, FieldDescriptor(field_name))

    def _detect_multipart(self, should_detect):
        if should_detect:
            for field_name, field in self.fields.items():
                if isinstance(field, File):
                    self.attributes['enctype'] = 'multipart/form-data'

    def _set_values_provider(self, values_provider):
        for field_name, field in self.fields.items():
            if hasattr(values_provider, field_name):
                field.values = getattr(values_provider, field_name)

    # field methods

    @cached_property
    def fields(self):
        self._mapped_fields = {}
        fields = collections.OrderedDict()
        for field_name, field in self._defined_fields.items():
            instance = field.generate_instance(self)
            setattr(self.__class__, field_name, FieldDescriptor(field_name))
            if not instance.name:
                instance.name = field_name
            self._mapped_fields[instance.name] = instance
            fields[field_name] = instance
        return fields

    @property
    def defined_fields(self):
        return self._defined_fields

    # data methods

    @cached_property
    def raw_data(self):
        """Returns a dict containing all the original field values.

        Field values will be their pre-filtered values.
        """
        return {field_name: field.original_value for
                field_name, field in self.fields.items()}

    @property
    def data(self):
        """Returns a dict containing all the field values.

        Used as a shorthand method to retrieve data from all the form fields
        rather than having to access the fields themselves.
        """
        return {field_name: field.value for
                field_name, field in self.fields.items()}

    @data.setter
    def data(self, data):
        """Sets the data for the form.

        Iterates through all the fields on the form and injects the value.

        Args:
            data (dict|watson.http.messages.Request): A dict of key/value pairs to populate the form with.
        """
        self.invalidate()
        if hasattr(data, 'post'):
            raw_data = MultiDict()
            raw_data.update(data.post.items())
            for key, value in data.files.items():
                # need to do this rather than .update due to File objects
                # being named tuples
                raw_data[key] = value
        else:
            raw_data = data
        self._set_data_on_fields(raw_data)

    def _set_data_on_fields(self, data):
        # internal method for setting the data on the fields
        for key in self._mapped_fields:
            if key not in self._ignored_bound_fields:
                field = self._mapped_fields[key]
                field.value = data.get(key, field.default_value)

    # error methods

    @cached_property
    def errors(self):
        """Returns a list of errors associated with the form.

        If the form has not been validated yet, calling this property
        will cause validation to occur.
        """
        self.is_valid()
        errors = {}
        for field_name, field in self.fields.items():
            error_list = field.errors
            if error_list:
                errors[field_name] = {'messages': field.errors,
                                      'label': field.label.text}
        if self._form_errors:
            errors['form'] = {'messages': self._form_errors, 'label': 'Form'}
        return errors

    # binding methods

    @property
    def bound_object(self):
        return self._bound_object

    def bind(self, obj=None, mapping=None, ignored_fields=None, hydrate=True):
        """Binds an object to the form.

        Optionally additional mapping can be specified in order to set values on
        any of the classes that may exist within the object.
        If this method is called after the data has been set on the form, then
        the existing data will be overridden with the attributes on the object
        unless hydrate is set to false.

        Args:
            obj (class|dict): the class to bind to the form.
            mapping (dict): the mapping between the form fields and obj attributes.
            ignored_fields (list|tuple): fields to ignore when binding.
            bool hydrate: whether or not to hydrate the form with the obj attributes.

        Example:

        .. code-block:: python

            form = ...
            user = User(username='test')
            form.bind(user)
            form.username.value  # 'test'
        """
        self._ignored_bound_fields = ignored_fields or []
        if obj:
            self._bound_object = obj
            if mapping:
                self._bound_object_mapping = mapping
        self.invalidate()
        if obj and hydrate:
            self.__hydrate_obj_to_form()

    # validation methods

    @property
    def validated(self):
        return self._validated

    def invalidate(self):
        """Invalidate the data that has been bound on the form.

        This is called automatically when data is bound to the form and
        sets the forms validity to invalid.
        """
        attrs = ('_data', '_raw_data', '_errors', '_form_errors')
        for attr in attrs:
            with suppress(AttributeError):
                delattr(self, attr)
        self._validated = self._valid = False

    def is_valid(self):
        """Determine whether or not the form and relating values are valid.

        Filter all the values on the fields associated with the form, and
        then validate each field. Will only execute the filter/validation
        steps if the form has not been previously validated, or has
        been invalidated.

        Returns:
            boolean value depending on the validity of the form.
        """
        if not self.validated:
            self._form_errors = []
            self._valid = True
            for field_name, field in self.fields.items():
                field.filter()
                valid = field.validate()
                if len(valid) > 0:
                    self._valid = False
            if self._valid:
                for validator in self.validators:
                    try:
                        validator(self)
                    except ValueError as exc:
                        self._valid = False
                        self._form_errors.append(str(exc))
            self._validated = True
        if self._valid and self._bound_object:
            self.__hydrate_form_to_obj()
        return self._valid

    # rendering methods

    def open(self, **kwargs):
        """Render the start tag of the form.

        Any addition kwargs will be used within the attributes.
        """
        self._rendering = True
        attrs = self.attributes.copy()
        if kwargs:
            attrs.update(kwargs)
        return '<form {0}>'.format(flatten_attributes(attrs))

    def close(self, include_http_request=True):
        """Render the end tag of the form.

        If the form has the http_request_method input then include it in the
        tag by default.

        Args:
            include_http_request (boolean): Whether or not to include the HTTP_REQUEST_METHOD field
        """
        self._rendering = False
        tag = '{0}</form>'
        field = ''
        if include_http_request and 'http_request_method' in self.fields:
            field = str(self.fields['http_request_method'])
        return tag.format(field)

    def render(self, with_tag='div', with_label=True):
        """Output the entire form as a string.

        Called automatically by the __str__ method.

        Args:
            with_tag (string): the tag to be used to separate the elements.
            with_label (boolean): render each field with it's label.

        Returns:
            A string representation of the form.
        """
        html = '{open}{fields}{close}'
        fields = ['<{0}>{1}</{0}>'.format(with_tag,
                                          field.render_with_label() if with_label else str(field))
                  for field_name, field in self.fields.items()]
        return (
            html.format(
                open=self.open(),
                close=self.close(False),
                fields=''.join(fields))
        )

    # convenience methods

    @property
    def name(self):
        return self.attributes['name']

    @property
    def method(self):
        return self.attributes['method']

    @property
    def action(self):
        return self.attributes['action']

    @property
    def enctype(self):
        return self.attributes['enctype']

    def __len__(self):
        """Return the number of fields associated with the form.
        """
        return len(self.fields)

    def __repr__(self):
        return '<{0} name:{1} method:{2} action:{3} fields:{4}>'.format(
            get_qualified_name(self),
            self.name,
            self.method,
            self.action,
            len(self))

    # hydration and internal methods

    def __hydrate_obj_to_form(self):
        # should never be called externally. Triggered by bind.
        obj_mapping = self._bound_object_mapping or {}
        for field_name, field in self.fields.items():
            current_obj = self._bound_object
            attr = field_name
            if field_name in obj_mapping:
                attr = obj_mapping[field_name][-1]
                fields = obj_mapping[field_name][0:-1]
                for name in fields:
                    try:
                        current_obj = getattr(current_obj, name)
                    except:
                        raise AttributeError(
                            'Mapping for object does not match object structure.')
            if hasattr(current_obj, attr) and attr not in self._ignored_bound_fields:
                self.fields[field_name].value = getattr(current_obj, attr)

    def __hydrate_form_to_obj(self):
        # should never be called externally. Triggered by is_valid.
        obj_mapping = self._bound_object_mapping or {}
        for field_name in self.data:
            current_obj = self._bound_object
            value = self.fields[field_name].value
            attr = field_name
            if field_name in obj_mapping:
                attr = obj_mapping[field_name][-1]
                for name in obj_mapping[field_name][0:-1]:
                    try:
                        current_obj = getattr(current_obj, name)
                    except:
                        raise AttributeError(
                            'Mapping for object does not match object structure.')
            if hasattr(current_obj, attr) and attr not in self._ignored_bound_fields:
                try:
                    setattr(current_obj, attr, value or None)
                except:  # pragma: no cover
                    # something nasty happened here, the user should manage it
                    pass


class Multipart(Form):

    """Convenience class for forms that should be multipart/form-data.

    By default, the Form class will automatically detect whether or not
    a field is of type file, and convert it to multipart.
    """

    def __init__(self, name, method='post', action=None, **kwargs):
        kwargs['enctype'] = 'multipart/form-data'
        super(Multipart, self).__init__(name, method, action, **kwargs)



# -*- coding: utf-8 -*-
from uuid import uuid4
import hashlib
from watson.common.datastructures import MultiDict
from watson.form import fields
from watson import validators


def has_csrf(cls):
    """Adds csrf protection to the form.

    Adds a new field named 'csrf_token' to the form and overrides the
    set data method to retrieve the correct token.

    If a form is csrf protected then a session object must be passed to
    the __init__ method so that a token can be created (if not already).

    Example:

    .. code-block:: python

        @has_csrf
        class MyForm(Form):
            username = fields.Text(required=True)
            password = fields.Password(required=True)
    """
    class CsrfProtectedForm(cls):
        csrf_token = fields.Csrf()

        def __init__(self, name, method='post', action=None,
                     detect_multipart=True, session=None, **kwargs):
            """Inititalize the form and set some default attributes.

            Args:
                name (string): the name of the form
                method (string): the http method to use
                action (string): the url to submit the form to
                detect_multipart (boolean): automatically set multipart/form-data
                session (watson.http.session.StorageMixin): the session in which to store the token
            """
            if not session:
                raise ValueError(
                    'A session must be assigned to the form for validation.')
            super(
                CsrfProtectedForm,
                self).__init__(
                name,
                method,
                action,
                detect_multipart,
                **kwargs)
            token_name = '{0}_csrf_token'.format(self.name)
            if token_name not in session or not session.get(token_name):
                token = '{0}{1}{2}'.format(token_name, uuid4().hex, session.id)
                actual = hashlib.sha256(token.encode('utf-8')).hexdigest()
                session[token_name] = actual
            for validator in self.fields['csrf_token'].validators:
                if isinstance(validator, validators.Csrf):
                    validator.token = session[token_name]
            self.csrf_token = session[token_name]

        @property
        def data(self):
            # required for the setter override
            return super(CsrfProtectedForm, self).data

        @data.setter
        def data(self, data):
            """Sets the data for the form.

            Override the existing set data method and inject the csrf
            token into the form.
            """
            token_name = '{0}_csrf_token'.format(self.name)
            if hasattr(data, 'post'):
                raw_data = MultiDict()
                raw_data.update(data.files.items())
                for key, value in data.post.items():
                    if key.endswith('_csrf_token'):
                        raw_data['csrf_token'] = value
                    else:
                        raw_data[key] = value
            else:
                if token_name in data:
                    data['csrf_token'] = data[token_name]
                    del data[token_name]
                raw_data = data
            self._set_data_on_fields(raw_data)

        def close(self):
            """Render the end tag of the form.

            Automatically renders the csrf field within the form.
            """
            return (
                '{0}{1}'.format(
                    self.fields['csrf_token'],
                    super(
                        CsrfProtectedForm,
                        self).close()))

    return CsrfProtectedForm



# -*- coding: utf-8 -*-
from datetime import datetime
import itertools
from watson.common.imports import get_qualified_name
from watson.html.elements import TagMixin, flatten_attributes
from watson import validators, filters as filters_


class Definition(object):
    """Placeholder form element which allows for the creation of new form
    elements when the form is instantiated.


    """
    _counter = itertools.count()

    def __init__(self, class_, *args, **kwargs):
        self.count = next(Definition._counter)
        self.class_ = class_
        self.args = args
        self.kwargs = kwargs

    def generate_instance(self, form):
        cls = self.class_
        self.kwargs['form_'] = form
        return cls(definition=False, *self.args, **self.kwargs)


class Label(TagMixin):

    """A <label> tag which can be automatically included with fields.

    Attributes:
        html (string): the html used to render the label
        text (string): the text associated with the label
    """
    html = '<label {0}>{1}</label>'
    text = None

    def __init__(self, text, **kwargs):
        self.text = text
        super(Label, self).__init__(**kwargs)

    def render(self, field=None, **kwargs):
        attrs = self.attributes.copy()
        if 'text' in kwargs:
            self.text = kwargs['text']
            del kwargs['text']
        if 'for_' in kwargs:
            attrs['for'] = kwargs['for_']
            del kwargs['for_']
        attrs.update(kwargs)
        if field and 'id' not in field.attributes and field.name:
            # inject id based on field name
            id = field.name
            field.attributes['id'] = id
            attrs['for'] = id
        return self.html.format(flatten_attributes(attrs), self.text)

    __call__ = render


class FieldMixin(TagMixin):

    """A mixin that can be used as a base to simplify the creation of fields.

    When defining a field, a fully instantiated field must be created with
    definition=False as an argument in it's __init__ method. This is to
    facilitate the way fields are defined in Form objects in 2.0.0.

    Attributes:
        label (watson.form.fields.Label): the label associated with the field
        html (string): the html used to render the field
        validators (list): the validators that will be used to validate the
                           value.
        filters (list): the filters that will be used prior to validation
    """
    _counter = itertools.count()
    label = None
    html = '{0}'
    validators = None
    filters = None
    _errors = None
    _value = None
    _default_value = None
    _original_value = None

    def __new__(cls, definition=True, *args, **kwargs):
        if definition:
            return Definition(cls, *args, **kwargs)
        return super(FieldMixin, cls).__new__(cls)

    def __init__(self, name=None, value=None, default_value=None,
                 label=None, label_attrs=None, **kwargs):
        """Initializes the field with a specific name.
        """
        self.count = next(FieldMixin._counter)
        if not name:
            name = ''
        self.label = Label(label or name)
        if label_attrs and isinstance(label_attrs, dict):
            self.label.attributes.update(label_attrs)
        kwargs['name'] = name
        self.value = value
        self.default_value = default_value
        self.filters = [filters_.Trim()] + kwargs.get('filters', [])
        self.validators = kwargs.get('validators', [])
        if 'validators' in kwargs:
            del kwargs['validators']
        if 'filters' in kwargs:
            del kwargs['filters']
        if '_class' in kwargs:
            kwargs['class'] = kwargs.get('_class')
            del kwargs['_class']
        if 'required' in kwargs:
            self.validators.append(validators.Required())
            kwargs['required'] = 'required'
        if 'form_' in kwargs:
            self.form = kwargs['form_']
            del kwargs['form_']
        self._errors = []
        super(FieldMixin, self).__init__(**kwargs)

    @property
    def value(self):
        """Return the value for the field.

        If the field has been cleaned, the original value can be retrieved
        with FieldMixin.original_value.
        """
        return self._value

    @value.setter
    def value(self, value):
        """Convenience method to set the value on the field.
        """
        self._value = value

    @property
    def default_value(self):
        return self._default_value

    @default_value.setter
    def default_value(self, value):
        if not self.value:
            self.value = value

    @property
    def original_value(self):
        """Return the original value for the field.
        """
        return self._original_value if self._original_value else self.value

    def filter(self):
        """Filter the value on the field based on the associated filters.

        Set the original_value of the field to the first value stored. Note, if
        this is called a second time, then the original value will be
        overridden.
        """
        for _filter in self.filters:
            self._original_value = self.value
            self.value = _filter(self.value)

    def validate(self):
        """Validate the value of the field against the associated validators.

        Returns:
            A list of errors that have occurred when the field has been
            validated.
        """
        self._errors = []
        for validator in self.validators:
            try:
                validator(self.value)
            except ValueError as exc:
                self._errors.append(str(exc))
        return self._errors

    @property
    def errors(self):
        return self._errors

    @property
    def name(self):
        """Convenience method to retrieve the name of the field.
        """
        return self.attributes['name']

    @name.setter
    def name(self, name):
        """Override the name attribute on the field.
        """
        self.attributes['name'] = name
        if not self.label.text:
            self.label.text = name

    def render_with_label(self):
        """Render the field with the label attached.
        """
        raise NotImplementedError('The render method has not been implemented')

    def __str__(self):
        return self.render()

    def __call__(self, **kwargs):
        return self.render(**kwargs)

    def __repr__(self):
        return '<{0} name:{1}>'.format(get_qualified_name(self), self.name)


class Input(FieldMixin):

    """Creates an <input> field.

    Custom input types can be created by sending type='type' through the
    __init__ method.

    Example:

    .. code-block:: python

        input = Input(type='text')  # <input type="text" />
    """
    html = '<input {0} />'

    def render(self, **kwargs):
        """Render the element as html.

        Does not need to be called directly, as will be called by __str__
        natively.
        """
        attributes = self.attributes.copy()
        attributes.update(kwargs)
        if self.value:
            attributes['value'] = str(self.value)

        return self.html.format(flatten_attributes(attributes))

    def render_with_label(self, **kwargs):
        """Render the element as html and include the label.

        Output the element and prepend the <label> to it.
        """
        return ''.join((self.label.render(self), self.render(**kwargs)))


class GroupInputMixin(Input):

    """A mixin for form elements that are used in a group.

    Related form elements are wrapped in a fieldset, with a common legend.
    """
    label_position = 'left'
    wrapped = True
    values = None
    fieldset_html = '<fieldset><legend>{0}</legend>{1}</fieldset>'

    def __init__(self, name=None, values=None, value=None, **kwargs):
        if 'label_position' in kwargs:
            self.label_position = kwargs['label_position']
            del kwargs['label_position']
        super(GroupInputMixin, self).__init__(name, value, **kwargs)
        if not values:
            values = value
        try:
            iter(values)
            self.values = values
        except:
            self.values = [(self.label.text, values)]

    def has_multiple_elements(self):
        """Determine whether or not a field has multiple elements.
        """
        return isinstance(self.values, (tuple, list)) and len(self.values) > 1

    def render(self, **kwargs):
        multiple_elements = self.has_multiple_elements()
        elements = []
        id = self.attributes.get('id', self.name)
        for index, label_value_pair in enumerate(self.values):
            attributes = self.attributes.copy()
            label_text, value = label_value_pair
            if multiple_elements:
                element_id = '{0}_{1}'.format(id, index)
            else:
                element_id = id
            attributes.update({
                'name': self.name,
                'id': element_id
            })
            attributes.update(kwargs)
            if value:
                attributes['value'] = value
            value = str(value)
            if isinstance(self.value, (list, tuple)):
                if value in (str(val) for val in self.value):
                    attributes['checked'] = 'checked'
            elif self.value and value == str(self.value):
                attributes['checked'] = 'checked'
            flat_attributes = flatten_attributes(attributes)
            element = self.__render_input(
                element_id,
                flat_attributes,
                label_text)
            elements.append(element)
        return ''.join(elements)

    def render_with_label(self, **kwargs):
        multiple_elements = self.has_multiple_elements()
        if multiple_elements:
            wrap_html = self.fieldset_html
        else:
            wrap_html = '{1}'
        return wrap_html.format(self.label.text, self.render(**kwargs))

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if self.has_multiple_elements():
            value = [value]
        self._value = value

    def __render_input(self, id, attributes, label_text):
        element = self.html.format(attributes)
        output = '{0}{1}'
        attrs = self.label.attributes.copy()
        attrs['for'] = id
        flat_attrs = flatten_attributes(attrs)
        if self.wrapped:
            if self.label_position == 'left':
                return (
                    self.label.html.format(
                        flat_attrs,
                        output.format(
                            label_text,
                            element))
                )
            return (
                self.label.html.format(
                    flat_attrs,
                    output.format(element,
                                  label_text))
            )
        else:
            label = self.label.html.format(flat_attrs, label_text)
            if self.label_position == 'left':
                return output.format(label, element)
            return output.format(element, label)


class Radio(GroupInputMixin):

    """Creates a radio input.

    Example:

    .. code-block:: python

        field = Radio(name='test', label='My Radio Options', values=(('Test', 1), ('Testing', 2)))
        str(field)

    .. code-block:: html

        <fieldset>
            <legend>My Radio Options</legend>
            <label for="test_0">Test<input id="test_0" name="test" type="radio" value="1" /></label>
            <label for="test_1">Testing<input id="test_1" name="test" type="radio" value="2" /></label>
        </fieldset>

    .. code-block:: python

        field = Radio(name='test', label='My Radio', values=1)
        str(field)

    .. code-block:: html

        <label for="test"><input type="radio" name="test" values="1" />My Radio</label>
    """

    def __init__(self, name=None, values=None, value=None, **kwargs):
        """Initializes the radio.

        If a value is specified, then that value out of the available values will
        be checked.
        If multiple values are specified, then a radio group will be created.

        Args:
            string name: the name of the field
            tuple|list values: the values to be used
            mixed value: the value for the field
        """
        super(
            Radio,
            self).__init__(
            name,
            values,
            value,
            type='radio',
            **kwargs)


class Checkbox(GroupInputMixin):

    """Creates a checkbox input.

    Example:

    .. code-block:: python

        field = Checkbox(name='test', label='My Radio Options', values=(('Test', 1), ('Testing', 2)))
        str(field)

    .. code-block:: html

        <fieldset>
            <legend>My Checkbox Options</legend>
            <label for="test_0">Test<input id="test_0" name="test" type="checkbox" /></label>
            <label for="test_1">Testing<input id="test_1" name="test" type="checkbox" /></label>
        </fieldset>

    .. code-block:: python

        field = Checkbox(name='test', label='My Checkbox', values=1)
        str(field)=None

    .. code-block:: html

        <label for="test"><input type="checkbox" name="test" value="1" />My Checkbox</label>
    """

    def __init__(self, name=None, values=None, value=None, **kwargs):
        """Initializes the checkbox.

        If a value is specified, then that value out of the available values will
        be checked.
        If multiple values are specified, then a checkbox group will be created.

        Args:
            name (string): the name of the field
            values (tuple|list): the values to be used
            value (mixed): the value for the field
        """
        super(
            Checkbox,
            self).__init__(
            name,
            values,
            value,
            type='checkbox',
            **kwargs)


class Button(Input):

    """Creates a button, can be used instead of Input(type="button").
    """
    html = '<button {0}>{1}</button>'

    def render(self, **kwargs):
        attributes = self.attributes.copy()
        attributes.update(kwargs)
        if self.value:
            attributes['value'] = str(self.value)
        return (
            self.html.format(flatten_attributes(attributes), self.label.text)
        )

    def render_with_label(self, **kwargs):
        return self.render(**kwargs)


class Submit(Input):

    """Creates a submit input.

    Attributes:
        button_mode (bool): whether or not to render as <button> or <input>
    """
    button_mode = False

    def __init__(self, name=None, value=None, button_mode=False, **kwargs):
        real_value = value or kwargs.get('label', name)
        if button_mode:
            self.html = '<button {0}>{1}</button>'
            self.button_mode = True
        super(Submit, self).__init__(name, real_value, type='submit', **kwargs)

    def render(self, **kwargs):
        if self.button_mode:
            attributes = self.attributes.copy()
            attributes.update(kwargs)
            return (
                self.html.format(
                    flatten_attributes(attributes),
                    self.label.text)
            )
        return super(Submit, self).render()

    def render_with_label(self, **kwargs):
        return self.render(**kwargs)


class Textarea(Input):

    """Creates a textarea field.
    """
    html = '<textarea {0}>{1}</textarea>'

    def render(self, **kwargs):
        attributes = self.attributes.copy()
        attributes.update(kwargs)
        value = self.value if self.value else ''
        return self.html.format(flatten_attributes(attributes), value)


class Select(FieldMixin):

    """Creates a select field.

    Attributes:
        html (string): the html for the outer select element
        option_html (string): the individual option html element
        optgroup_html (string): the optgroup html element
        options (list|dict): the options available
    """
    html = '<select {0}>{1}</select>'
    option_html = '<option value="{0}"{2}>{1}</option>'
    optgroup_html = '<optgroup label="{0}">{1}</optgroup>'
    options = None

    def __init__(self, name=None, options=None,
                 value=None, multiple=False, **kwargs):
        """Initializes the select field.

        If the options passed through are a dict, and the value of each key is
        a list or tuple, then an optgroup will be rendered, using the key
        as the label for the optgroup.

        Args:
            name (string): the name of the field
            options (list|dict): the options available
            value (string): the selected value
            multiple (bool): whether or not to allow multiple selections

        Example:

        .. code-block:: python

            field = Select(name='test', options=collections.OrderedDict([('Group One', [1, 2]), ('Group Two', [1, 2])]))
            str(field)

        .. code-block:: html

            <select name="test">
                <opgroup label="Group One">
                    <option value="1">1</option>
                </optgroup>
                <opgroup label="Group Two">
                    <option value="2">2</option>
                </optgroup>
            </select>
        """
        self.options = options or []
        if multiple or isinstance(value, (tuple, list)):
            kwargs['multiple'] = 'multiple'
        super(Select, self).__init__(name, value, **kwargs)

    def render(self, **kwargs):
        attributes = self.attributes.copy()
        attributes.update(kwargs)
        return (
            self.html.format(
                flatten_attributes(attributes),
                self._options_render())
        )

    def render_with_label(self, **kwargs):
        return ''.join((self.label.render(self), self.render(**kwargs)))

    # Options can also be referenced as values
    @property
    def values(self):
        return self.options

    @values.setter
    def values(self, values):
        self.options = values

    def _options_render(self):
        # internal method the render the options
        if isinstance(self.options, dict):
            options = []
            for label, value in self.options.items():
                if isinstance(value, (tuple, list)):
                    options.append(
                        self.optgroup_html.format(label,
                                                  self.__render_options(value)))
                else:
                    options.append(self.__render_option(label, value))
            return ''.join(options)
        else:
            return self.__render_options(self.options)

    def __render_options(self, options):
        return (
            ''.join([self.__render_option(value, value) for value in options])
        )

    def __render_option(self, label, value):
        # internal method to render an individual option
        if isinstance(value, (tuple, list)):
            value, label = value
        match = False
        str_value = str(value)
        str_self_value = str(self.value)
        if str_value == str_self_value:
            match = True
        elif isinstance(self.value, (list, tuple)):
            str_values = (str(val) for val in self.value)
            match = str_value in str_values
        selected = ' selected="selected"' if match else ''
        return self.option_html.format(value, label, selected)

# Convenience classes for input types. Can use Input(type='something') instead
# if required to create a different input field.
# Some of the input types add additional validators and filters to simplify the
# process.


class Text(Input):

    """Creates an <input type="text" /> element.
    """

    def __init__(self, name=None, value=None, **kwargs):
        super(Text, self).__init__(name, value, type='text', **kwargs)


class Date(Input):

    """Creates an <input type="date" /> element.
    """

    format = None

    def __init__(self, name=None, value=None, format='%Y-%m-%d', **kwargs):
        self.format = format
        date_filter = filters_.Date(format)
        if 'filters' in kwargs:
            kwargs['filters'].append(date_filter)
        else:
            kwargs['filters'] = [date_filter]
        if format:
            self.format = format
        super(Date, self).__init__(name, value, type='date', **kwargs)

    def render(self, **kwargs):
        """Format the date in the format the HTML5 spec requires.
        """
        if self.value:
            if not isinstance(self.value, str):
                self.value = self.value.strftime(self.format)
            else:
                date = datetime.strptime(self.value, self.format)
                self.value = date.strftime(self.format)
        return super(Date, self).render(**kwargs)


class Email(Input):

    """Creates an <input type="email" /> element.
    """

    def __init__(self, name=None, value=None, **kwargs):
        super(Email, self).__init__(name, value, type='email', **kwargs)


class Hidden(Input):

    """Creates an <input type="hidden" /> element.
    """

    def __init__(self, name=None, value=None, **kwargs):
        super(Hidden, self).__init__(name, value, type='hidden', **kwargs)


class Csrf(Input):

    """Creates an <input type="hidden" /> element for use in csrf protection.
    """

    def __init__(self, name='csrf_token', value=None, **kwargs):
        super(
            Csrf,
            self).__init__(name,
                           value,
                           type='hidden',
                           label='Cross-Site Request Forgery',
                           required=True,
                           **kwargs)
        self.validators.append(validators.Csrf())


class Password(Input):

    """Creates an <input type="password" /> element.
    """

    def __init__(self, name=None, value=None, **kwargs):
        super(Password, self).__init__(name, value, type='password', **kwargs)


class File(Input):

    def __init__(self, name=None, value=None, **kwargs):
        super(File, self).__init__(name, value, type='file', **kwargs)

    def render(self):
        """Overridden to prevent value from being put back into the field.
        """
        attributes = self.attributes.copy()

        return self.html.format(flatten_attributes(attributes))
